import sys

from typing import Any, Callable, Mapping, Union, NamedTuple, List, Tuple
from importlib import import_module
from datetime import date, datetime
from pathlib import Path
from unittest.mock import patch

import yaml
from yaml.representer import Representer
from faker.providers import BaseProvider as FakerProvider

import snowfakery.data_gen_exceptions as exc
from snowfakery.utils.yaml_utils import SnowfakeryDumper

from numbers import Number


Scalar = Union[str, Number, date, datetime, None]
FieldDefinition = "snowfakery.data_generator_runtime_object_model.FieldDefinition"


class LineTracker(NamedTuple):
    filename: str
    line_num: int


class SnowfakeryPlugin:
    """Base class for all plugins.

    Currently plugins can declare new functions. Over time new superpowers
    for plugins will be introduced.

    Plugins generally look like this:

    class MyPlugin:
        class Functions: # must be named `Functions`.
            def func1(self, arg1, arg2, arg3=default):
                return something()

    This function would be invoked through the name `MyPlugin.func1()`

    Despite the name, the Functions namespace could also include
    constants or other namespaces, if that is useful to you.

    A RuntimeContext object is supplied on `self.context` with methods
    context.field_vars() and context.context_vars().

    context.field_vars() are the same field variables that can be used in
    templates.

    context.context_vars() is a mutable mapping contaiing
    values that are available to this object template and that of
    all children.

    context_vars are unique to each plugin and plugins do not see
    (or corrupt, or overwrite) context_vars for another plugin.

    Plugins can also keep internal state for global data, like any other
    Python object.
    """

    def __init__(self, interpreter):
        self.interpreter = interpreter
        self.context = PluginContext(self)

    def custom_functions(self, *args, **kwargs):
        """Instantiate, contextualize and return a function library

        Default behaviour is to return self.Function()."""
        functions = self.Functions()
        functions.context = self.context
        return functions

    def close(self, *args, **kwargs):
        pass


class ParserMacroPlugin:
    """Abstract base class for plugins that generate code.

    Note that these are currently undocumented and their
    interface is not necessarily stable."""

    pass


class PluginContext:
    "Exposes certain stable internals to plugins"

    def __init__(self, plugin):
        self.plugin = plugin
        self.interpreter = plugin.interpreter

    def field_vars(self):
        return self.interpreter.current_context.field_vars()

    def context_vars(self):
        return self.interpreter.current_context.context_vars(
            self.plugin.__class__.__name__
        )

    def evaluate_raw(self, field_definition):
        """Evaluate the contents of a field definition"""
        return field_definition.render(self.interpreter.current_context)

    def evaluate(self, field_definition):
        """Evaluate the contents of a field definition and simplify to a primitive value."""
        rc = self.evaluate_raw(field_definition)
        if isinstance(rc, Scalar.__args__):
            return rc
        elif hasattr(rc, "simplify"):
            return rc.simplify()
        else:
            raise f"Cannot simplify {field_definition}. Perhaps should have used evaluate_raw?"


def lazy(func: Any) -> Callable:
    """A lazy function is one that expects its arguments to be unparsed"""
    func.lazy = True
    return func


def resolve_plugins(
    plugin_specs: List[Tuple[str, object]], search_paths: List[Union[str, Path]]
):
    "Resolve a list of plugins and lineinfos"
    with plugin_path(search_paths):
        return [resolve_plugin(*plugin_spec) for plugin_spec in plugin_specs]


def plugin_path(search_paths):
    cwd_plugins = "./plugins"
    user_plugins = Path.home() / ".snowfakery/plugins"
    new_sys_path = [
        *sys.path,
        *(str(p) for p in search_paths),
        str(cwd_plugins),
        str(user_plugins),
    ]

    return patch.object(sys, "path", new_sys_path)


def resolve_plugin(plugin: str, lineinfo) -> object:
    """Resolve a plugin to a class"""
    cls = resolve_plugin_alternatives(plugin)
    if not cls:
        raise exc.DataGenImportError(
            f"Cannot find plugin: {plugin}", lineinfo.filename, lineinfo.line_num
        )

    if issubclass(cls, FakerProvider):
        return (FakerProvider, cls)
    elif issubclass(cls, SnowfakeryPlugin):
        return (SnowfakeryPlugin, cls)
    elif issubclass(cls, ParserMacroPlugin):
        return (ParserMacroPlugin, cls)
    else:
        raise exc.DataGenTypeError(
            f"{cls} is not a Faker Provider nor Snowfakery Plugin",
            lineinfo.filename,
            lineinfo.line_num,
        )


def resolve_plugin_alternatives(plugin):
    """Interpret the plugin declaration in 2 ways.
    Return it when we find one that matches.

    1. The new way, which allows shorter declarations.
    2. The old way, which required everything to be spelled out.

    In other words:

    - plugin: foo.bar.Baz

    Means.

    1. The new way:

    from foo.bar.Baz import Baz

    2. The old way:

    from foo.bar import Baz
    """
    prefix, class_name = plugin.rsplit(".", 1)
    testnames = [plugin + "." + class_name, plugin]

    for testname in testnames:
        module_name, class_name = testname.rsplit(".", 1)
        try:
            module = import_module(module_name)
            if hasattr(module, class_name):
                assert getattr(module, class_name)
                return getattr(module, class_name)
        except ModuleNotFoundError:
            pass


class PluginResult:
    """`PluginResult` objects expose a namespace that other code can access through dot-notation.

    PluginResults can be initialized with a dict or dict-like object.

    PluginResults are serialized to continuation files as dicts."""

    def __init__(self, result: Mapping):
        self.result = result

    def __getattr__(self, name):
        # ensures that it won't recurse
        return self.__dict__["result"][name]

    def __reduce__(self):
        return (self.__class__, (dict(self.result),))

    def __repr__(self):
        return f"<{self.__class__} {repr(self.result)}>"

    def __str__(self):
        return str(self.result)


# round-trip PluginResult objects through continuation YAML if needed.
SnowfakeryDumper.add_representer(PluginResult, Representer.represent_object)
yaml.SafeLoader.add_constructor(
    "tag:yaml.org,2002:python/object/apply:snowfakery.plugins.PluginResult",
    lambda loader, node: PluginResult(loader.construct_sequence(node)[0]),
)



from typing import NamedTuple


def add_after_statements(mappings):
    indexed_by_sobject = _index_by_sobject(mappings)

    for idx, (mapping_name, mapping) in enumerate(mappings.items()):
        for lookup in mapping.get("lookups", {}).values():
            target_table = lookup["table"]
            # PersonContacts are not real
            if target_table == "PersonContact":
                continue
            target_mapping_index = indexed_by_sobject[target_table]
            if target_mapping_index.first_instance >= idx:
                if not lookup.get("after"):
                    lookup["after"] = target_mapping_index.last_step_name


class MappingIndex(NamedTuple):
    first_instance: int
    last_step_name: str


def _index_by_sobject(mappings):
    indexed_by_sobject = {}
    for idx, (mapping_name, mapping) in enumerate(mappings.items()):
        # make an index of the order of objects
        sobject = mapping["sf_object"]
        existing_index = indexed_by_sobject.get(sobject)

        if existing_index:
            new_mi = MappingIndex(existing_index.first_instance, mapping_name)
        else:
            new_mi = MappingIndex(idx, mapping_name)
        indexed_by_sobject[sobject] = new_mi

    return indexed_by_sobject



from enum import Enum, auto

import yaml
from .utils.yaml_utils import SnowfakeryDumper

IdManager = "snowfakery.data_generator_runtime.IdManager"


class ObjectRow:
    """Represents a single row

    Uses __getattr__ so that the template evaluator can use dot-notation."""

    yaml_loader = yaml.SafeLoader
    yaml_dumper = SnowfakeryDumper
    yaml_tag = "!snowfakery_objectrow"

    __slots__ = ["_tablename", "_values", "_child_index"]

    def __init__(self, tablename, values=(), index=0):
        self._tablename = tablename
        self._values = values
        self._child_index = index

    def __getattr__(self, name):
        try:
            return self._values[name]
        except KeyError:
            raise AttributeError(name)

    def __str__(self):
        return str(self.id)

    def __repr__(self):
        try:
            return f"<ObjectRow {self._tablename} {self.id}>"
        except Exception:
            return super().__repr__()

    def __getstate__(self):
        """Get the state of this ObjectRow for serialization.

        Do not include related ObjectRows because circular
        references in serialization formats cause problems."""
        values = {k: v for k, v in self._values.items() if not isinstance(v, ObjectRow)}
        return {"_tablename": self._tablename, "_values": values}

    def __setstate__(self, state):
        for slot, value in state.items():
            setattr(self, slot, value)

    @property
    def _name(self):
        return self._values.get("name")


class ObjectReference(yaml.YAMLObject):
    def __init__(self, tablename, id):
        self._tablename = tablename
        self.id = id


class SlotState(Enum):
    """The current state of a NicknameSlot.

    UNUSED=empty, ALLOCATED=referenced, CONSUMED=object generated"""

    UNUSED = auto()
    ALLOCATED = auto()
    CONSUMED = auto()


class NicknameSlot(ObjectReference):
    """A slot that represents a Nickname or Tablename"""

    _tablename: str
    id_manager: IdManager
    allocated_id: int = None

    def __init__(self, tablename, id_manager):
        self._tablename = tablename
        self.id_manager = id_manager

    @property
    def id(self):
        "Get an id corresponding to this slot. Generate one if necessary."
        if self.allocated_id is None:
            self.allocated_id = self.id_manager.generate_id(self._tablename)
        return self.allocated_id

    def consume_slot(self):
        "Mark a slot as filled by an object and return its id for use by the object."
        rc = self.allocated_id
        self.allocated_id = SlotState.CONSUMED
        return rc

    @property
    def status(self):
        "Is the slot empty/unreferenced, allocated/referenced or consumed/used"
        if self.allocated_id is None:
            return SlotState.UNUSED
        elif isinstance(self.allocated_id, int):
            return SlotState.ALLOCATED
        elif self.allocated_id == SlotState.CONSUMED:
            return SlotState.CONSUMED

    def __repr__(self):
        return f"<NicknameSlot {self._tablename} {self.status} {self.allocated_id}>"



from snowfakery.plugins import ParserMacroPlugin
from snowfakery.data_generator_runtime_object_model import (
    ObjectTemplate,
    FieldFactory,
    SimpleValue,
    StructuredValue,
)
from snowfakery import data_gen_exceptions as exc


class Salesforce(ParserMacroPlugin):
    def SpecialObject(self, context, args) -> ObjectTemplate:
        """Currently there is only one special object defined: PersonContact"""
        sobj, nickname = self._parse_special_args(args)
        line_info = context.line_num()
        if sobj == "PersonContact":
            return self._render_person_contact(context, sobj, nickname, line_info)
        else:
            raise exc.DataGenError(
                f"Unknown special object '{sobj}'. Did you mean 'PersonContact'?",
                None,
                None,
            )

    def _render_person_contact(self, context, sobj, nickname, line_info):
        """Generate the code to render a person contact as CCI expects.

        Code generation is a better strategy for this than a runtime
        plugin because some analysis of the table structures happens
        at parse time.
        """
        fields = [
            FieldFactory(
                "IsPersonAccount",
                SimpleValue("true", **line_info),
                **line_info,
            ),
            FieldFactory(
                "AccountId",
                StructuredValue("reference", ["Account"], **line_info),
                **line_info,
            ),
        ]
        new_template = ObjectTemplate(
            sobj,
            filename=line_info["filename"],
            line_num=line_info["line_num"],
            nickname=nickname,
            fields=fields,
        )
        context.register_template(new_template)
        return new_template

    def _parse_special_args(self, args):
        """Parse args of SpecialObject"""
        nickname = None
        if isinstance(args, str):
            sobj = args
        elif isinstance(args, dict):
            sobj = args["name"]
            if not isinstance(sobj, str):
                raise exc.DataGenError(
                    f"`name` argument should be a string {sobj}: {type(sobj)}"
                )
            nickname = args.get("nickname")
            if nickname and not isinstance(nickname, str):
                raise exc.DataGenError(
                    f"`nickname` argument should be a string {nickname}: {type(sobj)}"
                )

        return sobj, nickname

    def ContentFile(self, context, args) -> ObjectTemplate:
        return {
            "Base64.encode": [
                {"File.file_data": {"encoding": "binary", "file": args.get("path")}}
            ]
        }



import importlib.util

from snowfakery.plugins import SnowfakeryPlugin


class Math(SnowfakeryPlugin):
    def custom_functions(self, *args, **kwargs):
        "Expose math functions to Snowfakery"
        mathmodule = importlib.util.find_spec("math")
        math = importlib.util.module_from_spec(mathmodule)

        math.round = round
        math.min = min
        math.max = max

        return math



from multiprocessing import Process, set_start_method
from threading import Thread
from time import time, sleep
from tempfile import TemporaryDirectory
from pathlib import Path
from collections import defaultdict

import click
from sqlalchemy import create_engine

from snowfakery import generate_data


@click.command()
@click.argument("recipe", type=click.Path())
@click.argument("num-records", type=int)
@click.argument("num-records-tablename", type=str)
@click.option("--number-of-processes", type=int, default=0)
@click.option("--number-of-threads", type=int, default=0)
def snowbench(
    num_records,
    num_records_tablename,
    recipe,
    number_of_processes=0,
    number_of_threads=0,
):
    """A benchmarking tool for Snowfakery and Snowfakery recipes.

    Note that Snowfakery itself is not multi-threaded or multi-process,
    but it takes only a few lines of code to wrap it in a parallel
    processing context. For example, it might be a few lines of
    shell script or 10 lines of Python code.

    The sweet spot for "number_of_processes" is usually the number of
    CPU cores you have. Processes are usually faster than threads for
    Snowfakery.

    In some weird circumstances number_of_processes=1 may behave
    differently than the default behaviour, because the single
    execution process is spawned off into a sub-process instead
    of being executed inline.
    """
    with TemporaryDirectory() as tempdir:
        start = time()
        Thread(daemon=True, target=status, args=[tempdir]).start()
        if not number_of_processes and not number_of_threads:
            print(
                "--number-of-processes and --number-of-threads not supplied.\n"
                "Using a single inline execution unit."
            )
            output_file = Path(tempdir) / "generated_data.db"
            snowfakery(recipe, num_records, num_records_tablename, output_file)
        else:
            multithreaded_test(
                recipe,
                num_records,
                num_records_tablename,
                tempdir,
                number_of_processes,
                number_of_threads,
            )
        report_databases(tempdir, start)


def multithreaded_test(
    recipe,
    num_records,
    num_records_tablename,
    tempdir,
    number_of_processes,
    number_of_threads,
):
    num_workers = number_of_processes + number_of_threads

    def make_args(tempdir, prefix, idx):
        output_file = f"{tempdir}/sqlite_{prefix}_{idx}.db"
        return recipe, num_records // num_workers, num_records_tablename, output_file

    processes = [
        Process(target=snowfakery, args=make_args(tempdir, "p", idx))
        for idx in range(number_of_processes)
    ]
    threads = [
        Thread(target=snowfakery, args=make_args(tempdir, "t", idx))
        for idx in range(number_of_threads)
    ]

    workers = threads + processes
    start = time()
    for worker in workers:
        worker.start()

    for worker in workers:
        worker.join()

    duration = time() - start
    return duration


def count_databases(tempdir):
    databases = Path(tempdir).glob("*.db")
    counts = defaultdict(int)
    for database in databases:
        count_database(database, counts)
    return dict(counts)


def status(tempdir):
    start = time()
    sleep(2)
    for i in range(1, 10 ** 20):
        print(".", end="", flush=True)
        sleep(1)
        if i % 60 == 0:
            print()
            report_databases(tempdir, start)
        elif i % 20 == 0:
            print()
            for f in Path(tempdir).glob("*.db"):
                print(f"{f} : {f.stat().st_size:,}")


def report_databases(tempdir, start_time):
    counts = count_databases(tempdir)
    for name, count in counts.items():
        print(name, count)
    total = sum(counts.values())
    duration = time() - start_time

    print(
        f"{total:,} records /",
        f"{int(duration):,} seconds =~",
        f"{int(total / duration):,}",
        "records per second",
        f"\n= {int((total / duration) * 3600):,}",
        "records per hour",
        f"\n= {int((total / duration) * 3600 *24):,}",
        "records per day",
    )


def count_database(filename, counts):
    dburl = f"sqlite:///{filename}?mode=ro"
    engine = create_engine(dburl)
    tables = engine.table_names()
    for table in tables:
        counts[table] += count_table(engine, table)
    return counts


def count_table(engine, tablename):
    return engine.execute(f"select count(Id) from '{tablename}'").first()[0]


def snowfakery(recipe, num_records, tablename, outputfile):
    assert Path(recipe).exists()
    output = f"sqlite:///{outputfile}"
    print(output)
    generate_data(
        recipe,
        target_number=(num_records, tablename),
        output_format="txt",
        dburl=output,
    )


def main():
    snowbench.main(prog_name="bench")


if __name__ == "__main__":  # pragma: no cover
    set_start_method("spawn")
    main()



from difflib import get_close_matches
from faker import Faker
import typing as T


class FakeNames(T.NamedTuple):
    f: Faker

    def Username(self):
        return f"{self.f.first_name()}_{self.f.last_name()}_{self.f.uuid4()}@{self.f.hostname()}"

    def Alias(self):
        return self.f.first_name()[0:8]

    def FirstName(self):
        return self.f.first_name()

    def LastName(self):
        return self.f.last_name()

    def Email(self):
        return self.f.ascii_safe_email()

    def RealisticMaybeRealEmail(self):
        return self.f.email()


class FakeData:
    """Wrapper Faker which adds Salesforce names and case insensitivity."""

    def __init__(self, faker: Faker):
        self.faker = faker
        self.fake_names = FakeNames(faker)

        # canonical form of names is lower-case
        self._lowers = {
            v.lower(): v for v in dir(self.fake_names) if not v.startswith("_")
        }

    def _get_fake_data(self, origname, *args, **kwargs):
        # faker names are all lower-case
        name = origname.lower()
        if name in self._lowers:
            meth = getattr(self.fake_names, self._lowers[name])
            return meth(*args, **kwargs)
        else:
            try:
                # TODO: look for FooBar as foo_bar
                return getattr(self.faker, name)(**kwargs)
            except AttributeError:
                pass

        msg = f"No fake data type named {origname}."
        match_list = get_close_matches(name, self._lowers.values(), n=1)
        if match_list:
            msg += f" Did you mean {match_list[0]}"
        raise AttributeError(msg)
